@@ -4,27 +4,32 @@ import (
 	"fmt"
 	"time"
 
-	"github.com/juanfont/headscale/hscontrol/policy"
 	"github.com/juanfont/headscale/hscontrol/types"
 	"github.com/samber/lo"
 	"tailscale.com/net/tsaddr"
 	"tailscale.com/tailcfg"
+	"tailscale.com/types/views"
 )
 
+// NodeCanHaveTagChecker is an interface for checking if a node can have a tag.
+type NodeCanHaveTagChecker interface {
+	NodeCanHaveTag(node types.NodeView, tag string) bool
+}
+
 func tailNodes(
-	nodes types.Nodes,
+	nodes views.Slice[types.NodeView],
 	capVer tailcfg.CapabilityVersion,
-	polMan policy.PolicyManager,
+	checker NodeCanHaveTagChecker,
 	primaryRouteFunc routeFilterFunc,
 	cfg *types.Config,
 ) ([]*tailcfg.Node, error) {
-	tNodes := make([]*tailcfg.Node, len(nodes))
+	tNodes := make([]*tailcfg.Node, 0, nodes.Len())
 
-	for index, node := range nodes {
-		node, err := tailNode(
+	for _, node := range nodes.All() {
+		tNode, err := tailNode(
 			node,
 			capVer,
-			polMan,
+			checker,
 			primaryRouteFunc,
 			cfg,
 		)
@@ -32,7 +37,7 @@ func tailNodes(
 			return nil, err
 		}
 
-		tNodes[index] = node
+		tNodes = append(tNodes, tNode)
 	}
 
 	return tNodes, nil
@@ -40,9 +45,9 @@ func tailNodes(
 
 // tailNode converts a Node into a Tailscale Node.
 func tailNode(
-	node *types.Node,
+	node types.NodeView,
 	capVer tailcfg.CapabilityVersion,
-	polMan policy.PolicyManager,
+	checker NodeCanHaveTagChecker,
 	primaryRouteFunc routeFilterFunc,
 	cfg *types.Config,
 ) (*tailcfg.Node, error) {
@@ -53,61 +58,64 @@ func tailNode(
 	// TODO(kradalby): legacyDERP was removed in tailscale/tailscale@2fc4455e6dd9ab7f879d4e2f7cffc2be81f14077
 	// and should be removed after 111 is the minimum capver.
 	var legacyDERP string
-	if node.Hostinfo != nil && node.Hostinfo.NetInfo != nil {
-		legacyDERP = fmt.Sprintf("127.3.3.40:%d", node.Hostinfo.NetInfo.PreferredDERP)
-		derp = node.Hostinfo.NetInfo.PreferredDERP
+	if node.Hostinfo().Valid() && node.Hostinfo().NetInfo().Valid() {
+		legacyDERP = fmt.Sprintf("127.3.3.40:%d", node.Hostinfo().NetInfo().PreferredDERP())
+		derp = node.Hostinfo().NetInfo().PreferredDERP()
 	} else {
 		legacyDERP = "127.3.3.40:0" // Zero means disconnected or unknown.
 	}
 
 	var keyExpiry time.Time
-	if node.Expiry != nil {
-		keyExpiry = *node.Expiry
+	if node.Expiry().Valid() {
+		keyExpiry = node.Expiry().Get()
 	} else {
 		keyExpiry = time.Time{}
 	}
 
 	hostname, err := node.GetFQDN(cfg.BaseDomain)
 	if err != nil {
-		return nil, fmt.Errorf("tailNode, failed to create FQDN: %s", err)
+		return nil, err
 	}
 
 	var tags []string
-	for _, tag := range node.RequestTags() {
-		if polMan.NodeCanHaveTag(node, tag) {
+	for _, tag := range node.RequestTagsSlice().All() {
+		if checker.NodeCanHaveTag(node, tag) {
 			tags = append(tags, tag)
 		}
 	}
-	tags = lo.Uniq(append(tags, node.ForcedTags...))
+	for _, tag := range node.ForcedTags().All() {
+		tags = append(tags, tag)
+	}
+	tags = lo.Uniq(tags)
 
-	routes := primaryRouteFunc(node.ID)
-	allowed := append(node.Prefixes(), routes...)
+	routes := primaryRouteFunc(node.ID())
+	allowed := append(addrs, routes...)
 	allowed = append(allowed, node.ExitRoutes()...)
 	tsaddr.SortPrefixes(allowed)
 
 	tNode := tailcfg.Node{
-		ID:       tailcfg.NodeID(node.ID), // this is the actual ID
-		StableID: node.ID.StableID(),
+		ID:       tailcfg.NodeID(node.ID()), // this is the actual ID
+		StableID: node.ID().StableID(),
 		Name:     hostname,
 		Cap:      capVer,
 
-		User: tailcfg.UserID(node.UserID),
+		User: tailcfg.UserID(node.UserID()),
 
-		Key:       node.NodeKey,
+		Key:       node.NodeKey(),
 		KeyExpiry: keyExpiry.UTC(),
 
-		Machine:          node.MachineKey,
-		DiscoKey:         node.DiscoKey,
+		Machine:          node.MachineKey(),
+		DiscoKey:         node.DiscoKey(),
 		Addresses:        addrs,
 		PrimaryRoutes:    routes,
 		AllowedIPs:       allowed,
-		Endpoints:        node.Endpoints,
+		Endpoints:        node.Endpoints().AsSlice(),
 		HomeDERP:         derp,
 		LegacyDERPString: legacyDERP,
-		Hostinfo:         node.Hostinfo.View(),
-		Created:          node.CreatedAt.UTC(),
+		Hostinfo:         node.Hostinfo(),
+		Created:          node.CreatedAt().UTC(),
 
-		Online: node.IsOnline,
+		Online: node.IsOnline().Clone(),
 
 		Tags: tags,
 
@@ -125,10 +133,12 @@ func tailNode(
 		tNode.CapMap[tailcfg.NodeAttrRandomizeClientPort] = []tailcfg.RawMessage{}
 	}
 
-	if node.IsOnline == nil || !*node.IsOnline {
-		// LastSeen is only set when node is
-		// not connected to the control server.
-		tNode.LastSeen = node.LastSeen
+	// Set LastSeen only for offline nodes to avoid confusing Tailscale clients
+	// during rapid reconnection cycles. Online nodes should not have LastSeen set
+	// as this can make clients interpret them as "not online" despite Online=true.
+	if node.LastSeen().Valid() && node.IsOnline().Valid() && !node.IsOnline().Get() {
+		lastSeen := node.LastSeen().Get()
+		tNode.LastSeen = &lastSeen
 	}
 
 	return &tNode, nil
