@@ -60,7 +60,6 @@ func TestNodeCanApproveRoute(t *testing.T) {
 		route      netip.Prefix
 		policy     string
 		canApprove bool
-		skipV1     bool
 	}{
 		{
 			name:  "allow-all-routes-for-admin-user",
@@ -766,10 +765,33 @@ func TestNodeCanApproveRoute(t *testing.T) {
 			canApprove: false,
 		},
 		{
-			name:  "empty-policy",
+			name:       "empty-policy",
+			node:       normalNode,
+			route:      p("192.168.1.0/24"),
+			policy:     `{"acls":[{"action":"accept","src":["*"],"dst":["*:*"]}]}`,
+			canApprove: false,
+		},
+		{
+			name:  "policy-without-autoApprovers-section",
 			node:  normalNode,
-			route: p("192.168.1.0/24"),
-			policy: `{"acls":[{"action":"accept","src":["*"],"dst":["*:*"]}]}`,
+			route: p("10.33.0.0/16"),
+			policy: `{
+				"groups": {
+					"group:admin": ["user1@"]
+				},
+				"acls": [
+					{
+						"action": "accept",
+						"src": ["group:admin"],
+						"dst": ["group:admin:*"]
+					},
+					{
+						"action": "accept",
+						"src": ["group:admin"],
+						"dst": ["10.33.0.0/16:*"]
+					}
+				]
+			}`,
 			canApprove: false,
 		},
 	}
@@ -777,7 +799,7 @@ func TestNodeCanApproveRoute(t *testing.T) {
 	for _, tt := range tests {
 		t.Run(tt.name, func(t *testing.T) {
 			// Initialize all policy manager implementations
-			policyManagers, err := PolicyManagersForTest([]byte(tt.policy), users, types.Nodes{&tt.node})
+			policyManagers, err := PolicyManagersForTest([]byte(tt.policy), users, types.Nodes{&tt.node}.ViewSlice())
 			if tt.name == "empty policy" {
 				// We expect this one to have a valid but empty policy
 				require.NoError(t, err)
@@ -789,14 +811,8 @@ func TestNodeCanApproveRoute(t *testing.T) {
 			}
 
 			for i, pm := range policyManagers {
-				versionNum := i + 1
-				if versionNum == 1 && tt.skipV1 {
-					// Skip V1 policy manager for specific tests
-					continue
-				}
-
-				t.Run(fmt.Sprintf("PolicyV%d", versionNum), func(t *testing.T) {
-					result := pm.NodeCanApproveRoute(&tt.node, tt.route)
+				t.Run(fmt.Sprintf("policy-index%d", i), func(t *testing.T) {
+					result := pm.NodeCanApproveRoute(tt.node.View(), tt.route)
 
 					if diff := cmp.Diff(tt.canApprove, result); diff != "" {
 						t.Errorf("NodeCanApproveRoute() mismatch (-want +got):\n%s", diff)
