@@ -1,8 +1,6 @@
 package templates
 
 import (
-	"fmt"
-
 	"github.com/chasefleming/elem-go"
 	"github.com/chasefleming/elem-go/attrs"
 )
@@ -31,7 +29,7 @@ func Windows(url string) *elem.Element {
 			),
 			elem.Pre(nil,
 				elem.Code(nil,
-					elem.Text(fmt.Sprintf(`tailscale login --login-server %s`, url)),
+					elem.Text("tailscale login --login-server "+url),
 				),
 			),
 		),
